package policyauthor;

import java.awt.Color;
import java.awt.event.*;

import javax.script.ScriptException;
import javax.swing.*;

import edu.wpi.cetask.TaskEngine;

/**
 * JPanel for setting the values in an Access Control Matrix (ACMatrix).
 * 
 * @author Paul Freitas
 */
public class ACMatrixPanel extends JPanel implements ActionListener {

	/**
	 * 
	 */
	private static final long serialVersionUID = 961915437647614184L;

	private String actionNames[];

	private String roleNames[];

	private ACMatrixCell guiMatrix[][];

	private JPanel actionLabels[];

	private JPanel roleLabels[];

	private MatrixObserver obs;

	/**
	 * Default Constructor.  No roles, No actions, no observer.
	 */
	public ACMatrixPanel() {
		actionNames = new String[0];
		roleNames = new String[0];
		initializePanel();
		obs = null;
	}

	/**
	 * No roles, no actions.
	 * @param o the observer for this panel.
	 */
	public ACMatrixPanel(MatrixObserver o) {
		actionNames = new String[0];
		roleNames = new String[0];
		initializePanel();
		obs = o;
	}

	/**
	 * Constructor with no observer.
	 * @param actions the actions to represent in the ACMatrix
	 * @param roles the roles to represent in the ACMatrix
	 */
	public ACMatrixPanel(String roles[],String actions[]) {
		actionNames = new String[actions.length];
		System.arraycopy(actions, 0, actionNames, 0, actions.length);
		roleNames = new String[roles.length];
		System.arraycopy(roles, 0, roleNames, 0, roles.length);
		initializePanel();
		obs = null;
	}

	/**
	 * SuperConstructor! Sets everything.
	 * @param actions the actions to represent in the ACMatrix
	 * @param roles the roles to represent in the ACMatrix
	 * @param o the observer to be notified when values in the ACMatrix get changed.
	 */
	public ACMatrixPanel(String roles[], String actions[], MatrixObserver o) {
		actionNames = new String[actions.length];
		System.arraycopy(actions, 0, actionNames, 0, actions.length);
		roleNames = new String[roles.length];
		System.arraycopy(roles, 0, roleNames, 0, roles.length);
		initializePanel();
		obs = o;
	}

	/**
	 * Create the actual widget.  Must be called every time the actions or roles are changed.
	 */
	private void initializePanel() {
		this.setLayout(new SpringLayout());
		this.add(new JLabel(""));

		guiMatrix = new ACMatrixCell[roleNames.length][];
		for (int i = 0; i < guiMatrix.length; i++) {
			guiMatrix[i] = new ACMatrixCell[actionNames.length];
		}
		actionLabels = new JPanel[actionNames.length];
		roleLabels = new JPanel[roleNames.length];

		ACMatrixCell nextMatrixSquare;
		JPanel nextLabel;
		for (int i = 0; i < actionNames.length; i++) {
			nextLabel = new JPanel();
			nextLabel.add(new RotatedLabel(actionNames[i]));
			this.add(nextLabel);
			actionLabels[i] = nextLabel;
		}
		for (int i = 0; i < roleNames.length; i++) {
			nextLabel = new JPanel();
			nextLabel.add(new JLabel(roleNames[i]));
			this.add(nextLabel);
			roleLabels[i] = nextLabel;
			for (int j = 0; j < actionNames.length; j++) {
				JMenuBar jmb = new JMenuBar();
				nextMatrixSquare = new ACMatrixCell(roleNames[i],
						actionNames[j], this, jmb);
				nextMatrixSquare.addActionListener(this);
				jmb.add(nextMatrixSquare);
				this.add(jmb);
				guiMatrix[i][j] = nextMatrixSquare;
			}
		}

		SpringUtilities.makeCompactGrid(this, roleNames.length + 1 // numrows
				, actionNames.length + 1 // numcols
				, 5, 5 // initX, initY
				, 5, 5); // padding
	}

	/**
	 * Get the index of the specified item in the given list.
	 * @param item the string to find
	 * @param itemlist the array it should be in
	 * @return the index in itemlist that the item can be found at
	 * @throws NotInMatrixException if the specified item could not be found.
	 */
	private int getIndex(String item, String[] itemlist)
			throws NotInMatrixException {
		for (int i = 0; i < itemlist.length; i++) {
			if (itemlist[i].equals(item)) {
				return i;
			}
		}

		if (itemlist == roleNames) {
			throw new NotInMatrixException(NotInMatrixException.ROLE, item);
		}

		throw new NotInMatrixException(NotInMatrixException.ACTION, item);
	}
	
	/**
	 * Highlight a cell
	 * @param role
	 * @param action
	 */
	public void highlightCell(String role, String action) {
		int ri = getIndex(role,roleNames);
		int ai = getIndex(action,actionNames);
		
		highlightCell(ri,ai);
	}
	
	/**
	 * Highlight a cell
	 * @param ri the index of the role
	 * @param ai the index of the action
	 */
	public void highlightCell(int ri,int ai) {
		highlightCell(ri,ai,ACMatrixCell.HIGHBACKGROUND);
	}
	
	/**
	 * Highlight a cell
	 * @param role
	 * @param action
	 * @param hcolor the color to set the cell to
	 */
	public void highlightCell(String role, String action, Color hcolor) {
		int ri = getIndex(role,roleNames);
		int ai = getIndex(action,actionNames);
		
		highlightCell(ri,ai,hcolor);
	}
	
	/**
	 * Highlight a cell
	 * @param ri the index of the role
	 * @param ai the index of the action
	 * @param hcolor the color to set the cell to
	 */
	public void highlightCell(int ri,int ai, Color hcolor) {
		guiMatrix[ri][ai].setHighlight(hcolor);
		this.repaint();
	}

	/**
	 * Highlight a row in the matrix.
	 * @param role the role whose row should be highlighted
	 */
	public void highlightRow(String role) {
		int ri = getIndex(role, roleNames);

		highlightRow(ri);
	}

	/**
	 * Highlight a row in the matrix.
	 * @param ri the index of the row to highlight
	 */
	public void highlightRow(int ri) {
		roleLabels[ri].setBackground(ACMatrixCell.HIGHBACKGROUND);
		for (int i = 0; i < guiMatrix[ri].length; i++) {
			guiMatrix[ri][i].setHighlight(true);
		}
		this.repaint();
	}

	/**
	 * Highlight a row in the matrix.
	 * @param role the role whose row should be highlighted
	 * @param hcolor the color to set the row to
	 */
	public void highlightRow(String role,Color hcolor) {
		int ri = getIndex(role, roleNames);

		highlightRow(ri,hcolor);
	}

	/**
	 * Highlight a row in the matrix.
	 * @param ri the index of the row to highlight
	 * @param hcolor the color to set the row to
	 */
	public void highlightRow(int ri,Color hcolor) {
		roleLabels[ri].setBackground(ACMatrixCell.HIGHBACKGROUND);
		for (int i = 0; i < guiMatrix[ri].length; i++) {
			guiMatrix[ri][i].setHighlight(hcolor);
		}
		this.repaint();
	}
	
	/**
	 * Unhighlight a row in the matrix.
	 * @param role the role whose row should not be highlighted
	 */
	public void unhighlightRow(String role) {
		int ri = getIndex(role, roleNames);

		unhighlightRow(ri);
	}

	/**
	 * Unhighlight a row in the matrix.
	 * @param ri the index of the row to unhighlight
	 */
	public void unhighlightRow(int ri) {
		roleLabels[ri].setBackground(this.getBackground());
		for (int i = 0; i < guiMatrix[ri].length; i++) {
			guiMatrix[ri][i].setHighlight(false);
		}
		this.repaint();
	}

	/**
	 * Highlight a column in the matrix.
	 * @param action the action whose column should be highlighted
	 */
	public void highlightCol(String action) {
		int ai = getIndex(action, actionNames);

		highlightCol(ai);
	}

	/**
	 * Highlight a column in the matrix.
	 * @param ai the index of the column to highlight
	 */
	public void highlightCol(int ai) {
		actionLabels[ai].setBackground(ACMatrixCell.HIGHBACKGROUND);
		for (int i = 0; i < guiMatrix.length; i++) {
			guiMatrix[i][ai].setHighlight(true);
		}
		this.repaint();
	}
	
	/**
	 * Highlight a column in the matrix.
	 * @param action the action whose column should be highlighted
	 * @param hcolor the color to set the column to
	 */
	public void highlightCol(String action, Color hcolor) {
		int ai = getIndex(action, actionNames);

		highlightCol(ai,hcolor);
	}

	/**
	 * Highlight a column in the matrix.
	 * @param ai the index of the column to highlight
	 * @param hcolor the color to set the column to
	 */
	public void highlightCol(int ai, Color hcolor) {
		actionLabels[ai].setBackground(ACMatrixCell.HIGHBACKGROUND);
		for (int i = 0; i < guiMatrix.length; i++) {
			guiMatrix[i][ai].setHighlight(hcolor);
		}
		this.repaint();
	}

	/**
	 * Unhighlight a column in the matrix.
	 * @param action the action whose column should not be highlighted.
	 */
	public void unhighlightCol(String action) {
		int ai = getIndex(action, actionNames);

		unhighlightCol(ai);
	}

	/**
	 * Unhighlight a column in the matrix.
	 * @param ai the index of the column to unhighlight.
	 */
	public void unhighlightCol(int ai) {
		actionLabels[ai].setBackground(this.getBackground());
		for (int i = 0; i < guiMatrix.length; i++) {
			guiMatrix[i][ai].setHighlight(false);
		}
		this.repaint();
	}

	/**
	 * Unhighlight everything in the matrix
	 */
	public void unhighlightAll() {
		for(int i = 0; i < guiMatrix.length; i++) {
			roleLabels[i].setBackground(this.getBackground());
			for(int j = 0; j < guiMatrix[i].length; j++) {
				guiMatrix[i][j].setHighlight(false);
			}
		}
		
		for(int i = 0; i < actionLabels.length; i++) {
			actionLabels[i].setBackground(this.getBackground());
		}
	}

	/**
	 * Required method to implement ActionListener
	 * Catches events raised by the ACMatrixCells on this panel.
	 */
	public void actionPerformed(ActionEvent e) {
		if (obs != null) {
			if (e.getSource().getClass() == ACMatrixCellSelection.class) {
				ACMatrixCellSelection acmcs = (ACMatrixCellSelection) e
						.getSource();
				MatrixObserver.ACDecision dec = null;
				switch (acmcs.getMDecision()) {

				case ACMatrixCellSelection.NOTAPPLICABLE:
					dec = MatrixObserver.ACDecision.NOTAPPLICABLE;
					break;

				case ACMatrixCellSelection.PERMIT:
					dec = MatrixObserver.ACDecision.PERMIT;
					break;

				case ACMatrixCellSelection.DENY:
					dec = MatrixObserver.ACDecision.DENY;
					break;

				}
				obs.update(acmcs.getMRole(), acmcs.getMAction(), dec);
				guiMatrix[getIndex(acmcs.getMRole(), roleNames)][getIndex(acmcs
						.getMAction(), actionNames)].setToDecision(acmcs
						.getMDecision());
			}
		}
	}

	/**
	 * Set the observer of this panel.
	 * @param obs
	 */
	public void setObs(MatrixObserver obs) {
		this.obs = obs;
	}

	/**
	 * @return the actions represented in the ACMatrix.
	 */
	public String[] getActionNames() {
		return actionNames;
	}

	/**
	 * Set a new list of actions for the ACMatrix to represent.
	 * Re-initializes the panel, losing all previously set decisions.
	 * @param actionNames
	 */
	public void setActionNames(String[] actionNames) {
		this.actionNames = actionNames;
		initializePanel();
	}

	/**
	 * @return the roles represented in the ACMatrix.
	 */
	public String[] getRoleNames() {
		return roleNames;
	}

	/**
	 * Set a new list of roles for the ACMatrix to represent.
	 * Re-initializes the panel, losing all previously set decisions.
	 * @param roleNames
	 */
	public void setRoleNames(String[] roleNames) {
		this.roleNames = roleNames;
		initializePanel();
	}
	
	/*public void copyJSMatrix(String retScript,TaskEngine e) {
	    try {
		e.evalGlobal(retScript + ".rownames","ACMatrixPanel.copyJSMatrix()");
		e.evalGlobal(retScript + ".colnames", "ACMatrixPanel.copyJSMatrix()");
	    } catch(ScriptException e) {
		e.
	    }
	}*/

	/**
	 * Main method which allows testing of this panel.
	 * @param args
	 */
	public static void main(String[] args) {
		JFrame frame = new JFrame("Matrix test");
		String roles[] = new String[3];
		roles[0] = "owner";
		roles[1] = "group";
		roles[2] = "all";

		String actions[] = new String[3];
		actions[0] = "read";
		actions[1] = "write";
		actions[2] = "execute";

		ACMatrixPanel ac = new ACMatrixPanel(roles, actions,
				new TestMatrixObserver());

		frame.getContentPane().add(ac);

		ac.highlightRow("owner");
		ac.highlightCell("group", "read",Color.RED);

		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.pack();
		frame.setVisible(true);
	}

	public void getMatrixVals(TaskEngine engine) throws ScriptException {
		for(int i = 0; i < roleNames.length; i++) {
			for(int j = 0; j < actionNames.length; j++) {
				boolean dec = (Boolean)engine.evalGlobal("$matrix.get('" + roleNames[i] + "','" + actionNames[j] + "') == ACDecision.PERMIT", "getMatrixVals");
				int intdec;
				if(dec) {
					intdec = ACMatrixCellSelection.PERMIT;
				} else {
					dec = (Boolean)engine.evalGlobal("$matrix.get('" + roleNames[i] + "','" + actionNames[j] + "') == ACDecision.DENY", "getMatrixVals");
					if(dec) {
						intdec = ACMatrixCellSelection.DENY;
					} else {
						intdec = ACMatrixCellSelection.NOTAPPLICABLE;
					}
				}

				guiMatrix[i][j].setToDecision(intdec);
			}
		}
	}

}
