package policyauthor.constraint;

import java.awt.Component;
import java.util.Iterator;
import java.util.LinkedList;

import javax.script.ScriptException;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.ListCellRenderer;

import policyauthor.constraint.ConstraintConfigurationPanel.VSListener;

import edu.wpi.cetask.TaskEngine;

/**
 * Java object for representing the arguments and print string of a JavaScript
 * Constraint. These are JavaScript objects defined in the script part of
 * policytaskmodel.xml
 * 
 * @author Paul Freitas
 * 
 */
public class ConstraintTemplate {

	private String jsType; // name of the object type in Javascript

	private int numargs; // total number of arguments referenced in the
	// format string.

	private String argnames[]; // all the UNIQUE names of the arguments
	private String vals[]; // all the values of the arguments
	private ArgType types[]; // the types of all the arguments (roles or
	// actions)

	private String formatStringPieces[]; // the segments of the Constraint's
	// formatstring which are not
	// arguments.
	private int fsargs[]; // the symbol table for the format string-if an
	// argument with a given name appears twice, this
	// will have one entry for each occurence, each
	// referring to the same index in argnames, vals,
	// and types
	private LinkedList<ArgObserver>[] argObservers; // observers for each
	// argument

	private TaskEngine e; // task engine where the JavaScript lives

	/**
	 * The types that any argument can have.
	 */
	public static enum ArgType {
		ROLE, ACTION
	};

	/**
	 * @param engine the engine where the JavaScript lives
	 */
	public ConstraintTemplate(TaskEngine engine) {
		e = engine;
	}
	
	/**
	 * @param jsType
	 *            the JavaScript constraint type
	 * @param engine
	 *            the engine where the JavaScript lives
	 */
	public ConstraintTemplate(String jsType, TaskEngine engine) {
		e = engine;
		setJSType(jsType);
	}
	
	/**
	 * Set the type and fields of this ConstraintTemplate to match a specific instance
	 * @param script a script that will return a JavaScript constraint
	 * @throws ScriptException 
	 */
	public void setToConstraintInstance(String script) throws ScriptException {
		//get type
		String jsType = (String)e.evalGlobal("(" + script + ").constraintType", "ConstraintTemplate setToConstraintInstance");
		setJSType(jsType);
		//set arguments
		for(int i = 0; i < argnames.length; i++) {
			vals[i] = (String)e.evalGlobal("(" + script + ")." + argnames[i], "ConstraintTemplate setToConstraintInstance: argument: " + argnames[i]);
		}
	}

	/**
	 * Set the value of an argument. If this is a new value, notifies all
	 * observers of that argument.
	 * 
	 * @param argNum
	 *            the argument number-corresponds to the order in which
	 *            arguments appear in the format string, counting duplicates.
	 * @param value
	 *            the value to set the argument to.
	 */
	public void setArg(int argNum, String value) {
		// it is the responsibility of other code to do type checking for this
		if (!value.equals(vals[fsargs[argNum]])) {
			vals[fsargs[argNum]] = value;
			notifyArgObservers(argNum);
		}
	}

	/**
	 * Get the value of an argument.
	 * 
	 * @param argNum
	 *            see argNum in setArg.
	 * @return the value of the specified argument.
	 */
	public String getArg(int argNum) {
		return vals[fsargs[argNum]];
	}

	/**
	 * @param argNum
	 *            see argNum in setArg.
	 * @return the type of the specified argument.
	 */
	public ArgType getArgType(int argNum) {
		return types[fsargs[argNum]];
	}

	/**
	 * @return the number of arguments in this constraint. Duplicates are
	 *         counted.
	 */
	public int getNumArgs() {
		return numargs;
	}

	/**
	 * @return the portions of the format string which are not arguments.
	 */
	public String[] getFormatStringPieces() {
		return formatStringPieces;
	}

	/**
	 * Change the type of this ConstraintTemplate. Erases all previous data.
	 * 
	 * @param jsType
	 *            the JavaScript type to set this ConstraintTemplate to.
	 */
	public void setJSType(String jsType) {
		this.jsType = jsType;
		String formatString;
		try {
			formatString = (String) (e.evalGlobal(jsType + ".templateString",
					"Guide"));
		} catch (ScriptException e) {
			throw new ConstraintException("Constraint type " + jsType
					+ " Unknown.");
		}

		parseFormatString(formatString);
	}

	// TODO: Document format string syntax.
	/**
	 * Parse a constraint's format string to populate the fields of this object.
	 * 
	 * @param formatString
	 *            the constraint's format string.
	 */
	private void parseFormatString(String formatString) {
		String[] fsp = formatString.split("%");

		formatStringPieces = new String[fsp.length / 2 + 1];
		numargs = fsp.length - formatStringPieces.length;
		argnames = new String[numargs];
		types = new ArgType[numargs];
		vals = new String[numargs];
		fsargs = new int[numargs];

		int i = 0;
		int piecei = 0;
		int argi = 0;

		while (i < fsp.length) {
			formatStringPieces[piecei] = fsp[i];
			i++;
			piecei++;
			if (i < fsp.length) {
				argnames[argi] = fsp[i];
				fsargs[argi] = argi;
				switch (fsp[i].charAt(0)) {
				case 'a':
					types[argi] = ArgType.ACTION;
					break;
				case 'r':
					types[argi] = ArgType.ROLE;
					break;
				default:
					throw new ConstraintException("Argument " + fsp[i]
							+ " does not have a recognized type.");
				}
				vals[argi] = "";
				i++;
				argi++;
			}
		}

		// remove duplicate arguments
		i = 0;
		boolean match;
		for (int fsai = 0; fsai < fsargs.length && i < argnames.length; fsai++) {
			match = false;
			for (int j = 0; j < i; j++) {
				if (argnames[i].equals(argnames[j])) {
					removeArg(i);
					fsargs[fsai] = j;
					j = i;
					match = true;
				}
			}
			if (!match) {
				i++;
			}
		}

		argObservers = new LinkedList[argnames.length];
		for (i = 0; i < argObservers.length; i++) {
			argObservers[i] = new LinkedList<ArgObserver>();
		}
	}

	/**
	 * Remove a duplicate argument from the arglist.
	 * 
	 * @param remidx
	 *            the index of the argument to remove.
	 */
	private void removeArg(int remidx) {
		String wan[] = new String[argnames.length - 1];
		ArgType wt[] = new ArgType[wan.length];
		String wv[] = new String[wan.length];

		for (int i = 0; i < remidx; i++) {
			wan[i] = argnames[i];
			wt[i] = types[i];
			wv[i] = vals[i];
		}
		for (int i = remidx; i < wan.length; i++) {
			wan[i] = argnames[i + 1];
			wt[i] = types[i + 1];
			wv[i] = vals[i + 1];
		}

		argnames = wan;
		types = wt;
		vals = wv;
	}

	/**
	 * @return the JavaScript constraint type this represents.
	 */
	public String getJSType() {
		return jsType;
	}

	/**
	 * @return JavaScript code to construct a new Constraint with the arguments
	 *         set in this object.
	 */
	public String getCreateScript() {
		String retstr = "new " + jsType + "(";
		for (int i = 0; i < argnames.length; i++) {
			if(i != 0) retstr += ",";
			retstr += "\"" + vals[i] + "\"";
		}
		retstr += ")";

		return retstr;
	}

	/**
	 * Attach an observer to an argument. That observer will be notified
	 * whenever the argument's value changes.
	 * 
	 * @param o
	 *            the argument observer to attach
	 * @param argNum
	 *            the argument the observer wants to observe
	 */
	public void attachArgObserver(ArgObserver o, int argNum) {
		argObservers[fsargs[argNum]].add(o);
	}

	/**
	 * Detach an observer from an argument.
	 * 
	 * @param o
	 *            the argument observer to detach
	 * @param argNum
	 *            the argument the observer is currently observing.
	 */
	public void detachArgObserver(ArgObserver o, int argNum) {
		argObservers[fsargs[argNum]].remove(o);
	}

	/**
	 * Let all the observers of a given argument know that it has been updated.
	 * 
	 * @param argNum
	 *            the argument number of the argument that was changed.
	 */
	private void notifyArgObservers(int argNum) {
		Iterator<ArgObserver> i = argObservers[fsargs[argNum]].iterator();
		while (i.hasNext()) {
			i.next().update(this, argNum, vals[fsargs[argNum]]);
		}
	}
	
	public String getPrintString() {
		String retval = "";

		for (int i = 0; i < numargs; i++) {
			retval += formatStringPieces[i];
			retval += getArg(i);
		}
		retval += formatStringPieces[formatStringPieces.length - 1];
		
		return retval;
	}

}
